// Quintessence
// Copyright 2006-2012 Ben Lincoln
// http://www.thelostworlds.net/
//

// This file is part of Quintessence.

// Quintessence is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// Quintessence is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with Quintessence (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Threading;

namespace Quintessence
{
    public partial class frmMain : Form
    {
        frmAbout aboutWindow;
        delegate void mDelegate(string nStatus);
        delegate void imageDelegate(Bitmap nImage);
        delegate void controlDelegate();

        //threads for opening/exporting files and getting the current texture
        protected Thread openFileThread;
        protected Thread exportThread;
        protected Thread getTextureThread;

        //path to the textures file
        protected string filePath;

        //is the texture file open?
        protected bool fileIsOpen;

        //does the display image need to be updated?
        protected bool imageUpdated;

        //buffer image for the picturebox
        protected Bitmap bufferImage;

        //length of the file in bytes
        protected ulong fileLength;

        //number of textures
        protected uint totalTextures;

        //current texture
        protected uint currentTexture;

        //path to export to
        protected string exportPath;

        //path to import from
        protected string importPath;

        //for doing things the thread-safe way
        protected string currentTextureText;
        protected int currentTextureIndex;

        public frmMain()
        {
            InitializeComponent();
            aboutWindow = new frmAbout();
            //getVQImage("Textures.VQ", 18432, 256, 256);
            //decode_twiddled_vq(pBm,data+2048,data,hdr->width,hdr->height,hdr->type[0]);
        }

        protected Bitmap getVQImage(long fileOffset)
        {
            int width = 256;
            int height = 256;
            FileStream fStream;
            BinaryReader bReader;
            ushort[] codeBook;
            byte[] imageData;

            fStream = new FileStream(filePath, FileMode.Open, FileAccess.Read);
            bReader = new BinaryReader(fStream);

            //read in the codebook
            fStream.Seek(fileOffset, SeekOrigin.Begin);
            codeBook = new ushort[1024];
            for (int i = 0; i < 1024; i++)
            {
                codeBook[i] = bReader.ReadUInt16();
            }
            //read in the image data
            imageData = new byte[128 * 128];
            for (int i = 0; i < imageData.GetUpperBound(0); i++)
            {
                imageData[i] = bReader.ReadByte();
            }

            bReader.Close();
            fStream.Close();
            return decodeTwiddledVQ(imageData, codeBook, 256, 256);
        }

        protected Bitmap decodeTwiddledVQ(byte[] imageData, ushort[] codeBook, int width, int height)
        {
            //this method is based on code from BERO
            Color[] codeBookTable;
            int[] twiddleTable;
            int x, y;
            int alpha;
            int aShift = 15;
            int rShift = 10;
            int gShift = 5;
            int bShift = 0;
            Bitmap decodedImage = new Bitmap(width, height);
            Color colour1, colour2, colour3, colour4;

            //initialize the twiddle table
            twiddleTable = initTwiddleTable();

            //read the codebook table values in from the raw data
            codeBookTable = new Color[4 * 256];
            for (x = 0; x < 1024; x++)
            {
                ushort c = codeBook[x];
                alpha = ((c >> aShift) & 1);
                if (alpha > 0)
                {
                    alpha = 255;
                }
                codeBookTable[x] = Color.FromArgb(alpha, ((c >> rShift) & 31) << 3, ((c >> gShift) & 31) << 3, ((c >> bShift) & 31) << 3); 
            }

            //decode the vq data based on the code book and detwiddle it using the table

            for (y = 0; y < height; y += 2)
            {
                for (x = 0; x < width; x += 2)
                {
                    int c = imageData[twiddleTable[y / 2] | (twiddleTable[x / 2] << 1)];
                    int baseEntry = c * 4;
                    colour1 = codeBookTable[baseEntry];
                    colour2 = codeBookTable[baseEntry + 1];
                    colour3 = codeBookTable[baseEntry + 2];
                    colour4 = codeBookTable[baseEntry + 3];
                    decodedImage.SetPixel(x, y, colour1);   //p3
                    decodedImage.SetPixel(x + 1, y, colour3);   //p4
                    decodedImage.SetPixel(x, y + 1, colour2);   // p1
                    decodedImage.SetPixel(x + 1, y + 1, colour4); // p2
                }
            }

            return decodedImage;
        }

        //    int decode_twiddled_vq(char *out,void *in0,void *in1,int width,int height,int mode)
        //{
        //    unsigned char vqtab[3*4*256];
        //    unsigned char *in;

        //    int twiddletab[1024];
        //    int x,y,wbyte;
        //    const SHIFTTBL *s;

        //    unsigned short *in_w = in1;
        //    char *p = vqtab;

        //    s = &shifttbl[mode];
        //    for(x=0;x<256*4;x++) {
        //        int c = in_w[x];
        //        p[R] = (c >> s->rshift) & s->rmask;
        //        p[G] = (c >> s->gshift) & s->gmask;
        //        p[B] = (c << s->bshift) & s->bmask;
        //        p+=3;
        //    }

        //    init_twiddletab(twiddletab);

        //    in = (unsigned char*)in0;
        //    wbyte = (width*3+3)&~3;
        //    for(y=0;y<height;y+=2) {
        //        char *p = out+(height-y-1)*wbyte;
        //        for(x=0;x<width;x+=2) {
        //            int c = in[twiddletab[y/2]|(twiddletab[x/2]<<1)];
        //            unsigned char *vq = &vqtab[c*12];

        //            p[0] = vq[0];
        //            p[1] = vq[1];
        //            p[2] = vq[2];

        //            p[3] = vq[6];
        //            p[4] = vq[7];
        //            p[5] = vq[8];

        //            p[0-wbyte] = vq[3];
        //            p[1-wbyte] = vq[4];
        //            p[2-wbyte] = vq[5];

        //            p[3-wbyte] = vq[9];
        //            p[4-wbyte] = vq[10];
        //            p[5-wbyte] = vq[11];

        //            p+=6;
        //        }
        //    }
        //}

        protected int[] initTwiddleTable()
        {
            //based on code by Marcus Tatest by way of BERO
            int[] twiddleTable;
            int x;

            twiddleTable = new int[1024];

            for (x = 0; x < 32; x++)
            {
                twiddleTable[x] = (x & 1) | ((x & 2) << 1) | ((x & 4) << 2) | ((x & 8) << 3) | ((x & 16) << 4);
            }
            for (x = 32; x < 1024; x++)
            {
                twiddleTable[x] = twiddleTable[x & 31] | (twiddleTable[(x >> 5) & 31] << 10);
            }

            return twiddleTable;
        }

        private void aboutQuintessenceToolStripMenuItem_Click(object sender, EventArgs e)
        {
            aboutWindow.Show();
        }

        //    /* form marcus tatest */
        //static void init_twiddletab(int *twiddletab)
        //{
        //    int x;
        //#if 0
        //  for(x=0; x<1024; x++)
        //    twiddletab[x] = (x&1)|((x&2)<<1)|((x&4)<<2)|((x&8)<<3)|((x&16)<<4)|
        //      ((x&32)<<5)|((x&64)<<6)|((x&128)<<7)|((x&256)<<8)|((x&512)<<9);
        //#else
        //    for(x=0;x<32;x++)
        //        twiddletab[x] = (x&1)|((x&2)<<1)|((x&4)<<2)|((x&8)<<3)|((x&16)<<4);
        //    for(;x<1024;x++)
        //        twiddletab[x] = twiddletab[x&31] | (twiddletab[(x>>5)&31]<<10);
        //#endif
        //}

        #region File IO

        protected void openFile()
        {
            mDelegate dcUpdateStatus = new mDelegate(updateStatus);
            this.Invoke(dcUpdateStatus, "Opening file");
            //file objects for IO
            FileStream fStream;
            BinaryReader bReader;
            //BinaryWriter bWriter;

            try
            {
                fStream = new FileStream(filePath, FileMode.Open);
                bReader = new BinaryReader(fStream);
                //bWriter = new BinaryWriter(fStream);
                bReader.Close();
                fStream.Close();
            }
            catch (IOException ioEx)
            {
                MessageBox.Show("There was an error opening the file. Make sure it is not in use.", "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }


            //get file length/number of textures
            FileInfo fInfo;
            fInfo = new FileInfo(filePath);
            fileLength = (ulong)fInfo.Length;

            //number of textures = (filelength - header(4096)) / (256 * 256 * 2)
            totalTextures = (uint)(fileLength - 18432) / (18432);

            updateTexDropdown();
            enableControls();
            this.Invoke(dcUpdateStatus, "Ready");


        }

        private void exportCurrent()
        {
            mDelegate dcUpdateStatus = new mDelegate(updateStatus);
            this.Invoke(dcUpdateStatus, "Exporting current texture");
            try
            {
                bufferImage.Save(exportPath, System.Drawing.Imaging.ImageFormat.Png);
            }
            catch (Exception ex)
            {
                MessageBox.Show("There was an error saving the file. Try a different location and/or filename.", "IO Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            this.Invoke(dcUpdateStatus, "Ready");
        }

        private void exportAll()
        {
            mDelegate dcUpdateStatus = new mDelegate(updateStatus);
            controlDelegate cDisable = new controlDelegate(disableControls);
            controlDelegate cEnable = new controlDelegate(enableControls);
            this.Invoke(cDisable);

            int iEA, jEA;
            Bitmap tempBitmap;

            for (iEA = 0; iEA < totalTextures; iEA++)
            {
                this.Invoke(dcUpdateStatus, "Exporting texture " + iEA + " of " + totalTextures);
                tempBitmap = getVQImage((long)(18432 + (iEA * 18432)));
                tempBitmap.Save(exportPath + "\\SRDCTex-" + zeroFill(iEA.ToString(), 4) + ".PNG", System.Drawing.Imaging.ImageFormat.Png);
            }
            this.Invoke(cEnable);
            this.Invoke(dcUpdateStatus, "Ready");
        }

        #endregion

        #region Image functions

        private void displayCurrentTexture()
        {
            mDelegate dcUpdateStatus = new mDelegate(updateStatus);
            this.Invoke(dcUpdateStatus, "Reading texture");
            //updateStatus("Reading texture: " + cboCurrentTexture.SelectedText);
            bufferImage = getVQImage((long)(18432 + (currentTextureIndex * 18432)));
            imageDelegate iUpdate = new imageDelegate(UpdateImage);
            this.Invoke(iUpdate, bufferImage);
            //imageUpdated = true;
            //updateStatus("Ready");
            this.Invoke(dcUpdateStatus, "Ready");
        }

        protected void UpdateImage(Bitmap newImage)
        {
            imgDisplay.Image = newImage;
        }

        #endregion

        private void cmdGo_Click(object sender, System.EventArgs e)
        {
            //user is a dumbfuck
            if (cboCurrentTexture.SelectedIndex == -1)
            {
                return;
            }
            currentTextureText = cboCurrentTexture.SelectedText;
            currentTextureIndex = cboCurrentTexture.SelectedIndex;
            getTextureThread = new Thread(new ThreadStart(displayCurrentTexture));
            getTextureThread.Start();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                exportThread.Abort();
            }
            catch (Exception ex)
            {
                //do nothing
            }
            try
            {
                openFileThread.Abort();
            }
            catch (Exception ex)
            {
                //do nothing
            }
            try
            {
                getTextureThread.Abort();
            }
            catch (Exception ex)
            {
                //do nothing
            }
            Application.Exit();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog fDialogue;
            DialogResult result;
            fDialogue = new OpenFileDialog();
            fDialogue.CheckFileExists = true;
            fDialogue.CheckPathExists = true;
            fDialogue.Multiselect = false;
            fDialogue.Title = "Select the Textures.VQ file to open...";
            result = fDialogue.ShowDialog();
            if (result == DialogResult.OK)
            {
                filePath = fDialogue.FileName;
                //openFileThread = new Thread(new ThreadStart(openFile));
                //openFileThread.Start();
                openFile();
            }
        }

        private void exportCurrentToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveDialogue;
            DialogResult result;
            saveDialogue = new SaveFileDialog();
            saveDialogue.AddExtension = true;
            saveDialogue.DefaultExt = "*.PNG";
            saveDialogue.FileName = "SRDCTex-" + zeroFill(cboCurrentTexture.SelectedIndex.ToString(), 4) + ".PNG";
            saveDialogue.Filter = "Portable Network Graphic Images (*.PNG)|*.PNG";
            result = saveDialogue.ShowDialog();
            if (result == DialogResult.OK)
            {
                exportPath = saveDialogue.FileName;
                exportThread = new Thread(new ThreadStart(exportCurrent));
                exportThread.Start();
                //exportCurrent();
            }
        }

        private void exportAllToolStripMenuItem_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog folderDialogue;
            DialogResult result;
            folderDialogue = new FolderBrowserDialog();
            folderDialogue.ShowNewFolderButton = true;
            result = folderDialogue.ShowDialog();
            if (result == DialogResult.OK)
            {
                exportPath = folderDialogue.SelectedPath;
                exportThread = new Thread(new ThreadStart(exportAll));
                exportThread.Start();
                //exportAll();
            }
        }

        private string zeroFill(string origVal, int length)
        {
            string retString;

            retString = origVal;

            do
                retString = "0" + retString;
            while (retString.Length < length);

            return retString;
        }

        public void updateStatus(string newStatus)
        {
            lblStatus.Text = "Status: " + newStatus;
        }

        private void updateTexDropdown()
        {
            int iUT;
            cboCurrentTexture.Items.Clear();
            for (iUT = 0; iUT <= totalTextures; iUT++)
            {
                cboCurrentTexture.Items.Add(iUT);
            }
            cboCurrentTexture.Text = "0";
        }

        private void enableControls()
        {
            exportToolStripMenuItem.Enabled = true;
            cmdGo.Enabled = true;
            openToolStripMenuItem.Enabled = true;
        }

        private void disableControls()
        {
            exportToolStripMenuItem.Enabled = false;
            cmdGo.Enabled = false;
            openToolStripMenuItem.Enabled = false;
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            updateStatus("Ready");
        }

    }


}